"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Keystore = void 0;
const debug_1 = __importDefault(require("debug"));
const errors_1 = require("./errors");
class Keystore {
    /**
     * Creates a new keystore service
     *
     * @param debugNamePrefix Name prefix used for the debug module
     * @param signingKeyPassphrases Stores passphrases for each signing keys. Key: key id, value: passphrase
     * @param keystoreReader Keystore reader callback (sync or async)
     */
    constructor(signingKeyPassphrases, keystoreReader, debugNamePrefix) {
        /** The current signing key ID */
        this.currentSigningKeyId = null;
        /** Cache for the private key and certificate data */
        this.keys = new Map();
        this.debug = debug_1.default(`${debugNamePrefix ? debugNamePrefix + ':' : ''}keystore`);
        this.signingKeyPassphrases = signingKeyPassphrases;
        this.keystoreReader = keystoreReader;
    }
    /**
     * Start reader task
     *
     * @param refreshInterval Interval of the keystore refresh [millisec]
     */
    async start(refreshInterval) {
        // first call before starting timer
        await this.keystoreReaderTask();
        this.interval = setInterval(this.keystoreReaderTask.bind(this), refreshInterval);
    }
    /**
     * Stop reader task
     */
    stop() {
        if (this.interval) {
            clearInterval(this.interval);
        }
    }
    /**
     * Selects the current signing key id
     *
     * @param currentKeys The current keys
     *
     * @return The selected key id
     */
    selectCurrentSigningKeyId(currentKeys) {
        if (currentKeys.size === 0) {
            return null;
        }
        if (currentKeys.size === 1) {
            return currentKeys.keys().next().value;
        }
        const now = Math.floor(Date.now() / 1000);
        const maxTimestamp = now - (20 * 60); // the youngest key to use
        const allKeys = Array.from(currentKeys);
        let allowedKeys = allKeys.filter(keyEntry => keyEntry[1].timestamp < maxTimestamp);
        this.debug('allowedKeys: ', allowedKeys.map(keyEntry => keyEntry[0]));
        if (allowedKeys.length === 0) {
            allowedKeys = allKeys;
        }
        allowedKeys = allowedKeys.sort((a, b) => a[0] > b[0] ? 1 : a[0] < b[0] ? -1 : 0);
        const entryToUse = allowedKeys[allowedKeys.length - 1];
        return entryToUse[0];
    }
    /**
     * Returns the ID of the signing key that has to be used for signing
     *
     * @return The ID of the key that has to be used for signing
     */
    getCurrentSigningKeyId() {
        return this.currentSigningKeyId;
    }
    /**
     * Returns the passphrase for the given private key
     *
     * @param id The id of the private key
     *
     * @return The passphrase for the private key
     */
    getPrivateKeyPassphrase(id) {
        if (this.signingKeyPassphrases.hasOwnProperty(id)) {
            return this.signingKeyPassphrases[id];
        }
        throw new errors_1.PrivateKeyPassphraseNotFoundError(id);
    }
    /**
     * Returns the private key with the given id
     * Throws PrivateKeyNotFoundError if the certificate is not found in the store
     *
     * @param id The private key id
     *
     * @return Promise to the private key {key(PEM), passphrase}
     */
    async getPrivateKey(id) {
        const key = this.keys.get(id);
        if (!key) {
            throw new errors_1.PrivateKeyNotFoundError(`Loading private key ${id} failed`);
        }
        return {
            alg: key.keyAlg,
            key: key.privKey,
            passphrase: this.getPrivateKeyPassphrase(id),
        };
    }
    /**
     * Returns the certificate with the given id
     * Throws CertificateNotFoundError if the certificate is not found in the store
     *
     * @param id The certificate id
     *
     * @return Promise to the certificate (PEM)
     */
    async getCertificate(id) {
        const key = this.keys.get(id);
        if (!key) {
            throw new errors_1.CertificateNotFoundError(`Loading certificate ${id} failed`);
        }
        return {
            alg: key.keyAlg,
            cert: key.cert,
        };
    }
    /**
     * Returns all certificates in JWKS format
     *
     * @return An array of JWK objects
     */
    getAllCertificatesAsJWKS() {
        const keySet = [];
        this.keys.forEach((key) => {
            keySet.push(key.pubkeyJwk);
        });
        return keySet;
    }
    async keystoreReaderTask() {
        try {
            const newKeys = await this.keystoreReader.readKeys(this.keys);
            this.debug('Keystore reloaded, keys: ', newKeys.keys());
            this.keys = newKeys;
            this.currentSigningKeyId = this.selectCurrentSigningKeyId(this.keys);
            this.debug('Current signing key id: ' + this.currentSigningKeyId);
        }
        catch (err) {
            // FIXME: should signal this error somehow. an EventEmitter maybe?
            this.debug('Reading keystore failed', err);
        }
    }
}
exports.Keystore = Keystore;
//# sourceMappingURL=Keystore.js.map